/* determines recursicely for all files in a directory and its subdirectories */
/* the block numbers and fragmentation */


#define VERSION "0.1"

#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64


#include <stdio.h>
#include <sys/ioctl.h>
#include <dirent.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <linux/fs.h>
#include <stdlib.h>
#include <unistd.h>


struct statistics            /* summary statistics */
{
  unsigned long n;           /* number of files in this class */
  unsigned long n_frag;      /* number of fragmented files */
  unsigned long frag;        /* number of fragments */
  unsigned long n_blocks;    /* number of blocks */
} global[6];


/* flags */
int recursive = 0;      /* 1: recurse into subdirectories */
int print_stats = 0;    /* 1: print statistics for each subdirectory */
int print_files = 0;    /* 1: print file fragmentation */
int print_blocks = 0;   /* 1: print block numbers as read; implies print_files */
int xdev = 0;           /* 1: ignore other file systems */



void usage(char *myname)
{
  printf("Usage: %s [-r] [-d] [-f] [-b] [-x] directory\n", myname);
  printf("-r: recurse into subdirectories\n");
  printf("-d: statistics for each subdirectory (implies -r)\n");
  printf("-f: print fragmentation of each file\n");
  printf("-b: print block numbers of each file processed (implies -f)\n");
  printf("-x: ignore other file systems when recursing\n");
  _exit(1);
}   /* usage() */



unsigned int scan_directory(char *dirname)
/* process directory dirname */
/* 0: error, else: number of block numbers in one block (n_indir) */
{
  DIR *d;                    /* dir handle to scan directory */
  struct dirent *dent;       /* directory entry */
  struct stat buf;           /* needed for lstat(dent) */
  char path[PATH_MAX];       /* complete path name of directory entry */
  int f;                     /* file descriptor */
  unsigned long b, b1;       /* block number */
  unsigned long n;           /* number of fragments in the file processed */
  unsigned long bsize = 0;   /* block size of file system */
  unsigned int n_indir = 0;  /* number of block numbers in one block */
  unsigned long n_blocks;    /* number of blocks of the file processed */
  struct statistics sta[6];  /* 0: 1 block; 1: <= 12 blocks; */
                             /* 2: <= 12+512; 3: <= 12+1024 */
                             /* 4: <= 12+4096; 5: > 12+4096 */
  struct statistics *st;     /* file class for statistics, depends on n_blocks */
  struct statistics sum;     /* summary statistic */
  unsigned long devnr;       /* device dirname is on */
  unsigned int i;
  
  for (i=0; i<6; i++)
  {
    sta[i].n = 0;
    sta[i].n_frag = 0;
    sta[i].frag = 0;
    sta[i].n_blocks = 0;
  }

  if (-1 == lstat(dirname, &buf))
  {
    printf(">>> Can't lstat() %s\n", path);
    return 0;
  }
  else devnr = buf.st_dev;
  
  d = opendir(dirname);
  if (!d)
  {
    printf(">>> can't opendir() %s\n", dirname);
    return 0;
  }
  
  while (dent = readdir(d))  /* read all directory entries */
  {
    if (1 == strlen(dent->d_name) && !strncmp(".", dent->d_name, 1)) continue;
    if (2 == strlen(dent->d_name) && !strncmp("..", dent->d_name, 2)) continue;
    sprintf(path, "%s/%s", dirname, dent->d_name);
    if (-1 == lstat(path, &buf))
    {
      printf(">>> Can't lstat() %s\n", path);
      continue;
    }

    if (S_ISLNK(buf.st_mode)) continue;
    else if (S_ISCHR(buf.st_mode)) continue;
    else if (S_ISBLK(buf.st_mode)) continue;
    else if (S_ISFIFO(buf.st_mode)) continue;
    else if (S_ISSOCK(buf.st_mode)) continue;
    else if (S_ISDIR(buf.st_mode))
    {
      if (recursive)
      {
        if (lstat(path, &buf) != -1)
        {
          if (devnr != buf.st_dev)
          {
            if (xdev)
            {
              printf(">>> skipping %s (different device)\n", path);
              continue;
            }
            else printf(">>> Warning: %s is a mounted device, use '-x' to skip", path);
          }
        }
        i = scan_directory(path); /* if we don't find a file in start dir */
        if (i) n_indir = i;       /* to determine n_indir, we take the last */
      }                           /* value we have. This avoids n_indir== 0 */
    }                             /* in case of last subdir being empty */

    else if (S_ISREG(buf.st_mode))
    {
      f = open(path, O_RDONLY|O_LARGEFILE);
      if (-1 == f) 
      {
        printf(">>> Can't open %s\n", path);
        continue;
      }

      if (!bsize) ioctl(f, FIGETBSZ, &bsize);
      if (!bsize)
      {
        printf(">>> ioctl(%s, FIGETBSZ) error\n", path);
        close(f);
        closedir(d);
        return 0;
      }
      n_blocks = (buf.st_size / bsize);
      if (buf.st_size % bsize > 0) n_blocks++;
      n_indir = bsize / sizeof (unsigned int);
      if (print_files)
       printf("%s (%lld bytes, %lu blocks):\n", path, (long long) buf.st_size, n_blocks);
      if (n_blocks <= 1) st = &sta[0];
      else if (n_blocks <= 12) st = &sta[1];
      else if (n_blocks <= 12+n_indir/2) st = &sta[2];
      else if (n_blocks <= 12+n_indir) st = &sta[3];
      else if (n_blocks <= 12+n_indir*4) st = &sta[4];
      else st = &sta[5];
      st->n++;
      st->n_blocks += n_blocks;

      b = 0;
      if (-1 == ioctl(f, FIBMAP, &b))  /* after call, b contains physical */
      {                                /* block number of bth block in file */
        printf(">>> ioctl(FIBMAP) error\n");
        close(f);
        continue;
      }

      n = 1;
      b1 = i = 1;
      while ( (ioctl(f, FIBMAP, &b1) != -1) && b1 )
      {       
        if (print_blocks) printf("%11lu", b);
        if (i % n_indir == 12)            /* indirect block */
        {
          if (i == 12+n_indir)         /* double indirect block (followed */
          {                            /* directly by an indirect block) */
            if (print_blocks) printf("        (DINDIR)         (INDIR)");
            if (b1-b != 3)
            {
              n++;
              if (print_blocks) printf(" *** ");
            }
            else if (print_blocks) printf("     ");
          }
          else
          {
            if (print_blocks) printf("         (INDIR)");
            if (b1-b != 2)
            {
              n++;
              if (print_blocks) printf(" *** ");
            }
            else if (print_blocks) printf("     ");
          }
        }   /* indirect block */
        else                      /* direct block */
        {
          if (b1-b != 1)
          {
            n++;
            if (print_blocks) printf(" *** ");
          }
          else if (print_blocks) printf("     ");
        }
        b = b1;
        i++;
        b1 = i;
      }   /* while(FIBMAP) */
      if (print_blocks) printf("%11lu\n", b);
      st->frag += n;
      if (n>1)
      {
        st->n_frag++;
        if (print_files) printf("%lu fragments\n", n);
      }
      else
      {
        if (print_files) printf("unfragmented\n");
      }
    }
    close(f);
  }   /* while(readdir()) */
  closedir(d);

  n = 0;
  for (i=0; i<6; i++) n += sta[i].n;   /* any files found? */
  if (n)
  {
    for (i=0; i<6; i++)
    {
      global[i].n += sta[i].n;
      global[i].n_frag += sta[i].n_frag;
      global[i].frag += sta[i].frag;
      global[i].n_blocks += sta[i].n_blocks;
    }
    if (print_stats)
    {
      printf("\nSummary statistics for %s:\n", dirname);
      printf(" blocks    files  fragmented  percent  fragm./file\n");
      printf("-----------------------------------------------------\n");
      printf("      1 %8lu           0     0.00         1.00\n", sta[0].n);
      printf("  <= 12 %8lu    %8lu %8.2f  %11.2f\n",
           sta[1].n, sta[1].n_frag, 
           sta[1].n > 0 ? sta[1].n_frag*100.0 / sta[1].n : 0, 
           sta[1].n > 0 ? sta[1].frag*1.0 / sta[1].n : 0);
      printf("<= %4u %8lu    %8lu %8.2f  %11.2f\n", 
           n_indir/2+12, sta[2].n, sta[2].n_frag, 
           sta[2].n > 0 ? sta[2].n_frag*100.0 / sta[2].n : 0,
           sta[2].n > 0 ? sta[2].frag*1.0 / sta[2].n : 0);
      printf("<= %4u %8lu    %8lu %8.2f  %11.2f\n", 
           n_indir+12, sta[3].n, sta[3].n_frag,
           sta[3].n > 0 ? sta[3].n_frag*100.0 / sta[3].n : 0,
           sta[3].n > 0 ? sta[3].frag*1.0 / sta[3].n : 0);
      printf("<= %4u %8lu    %8lu %8.2f  %11.2f\n",
           n_indir*4+12, sta[4].n, sta[4].n_frag,
           sta[4].n > 0 ? sta[4].n_frag*100.0 / sta[4].n : 0,
           sta[4].n > 0 ? sta[4].frag*1.0 / sta[4].n : 0);
      printf(" > %4u %8lu    %8lu %8.2f  %11.2f\n",
           n_indir*4+12, sta[5].n, sta[5].n_frag, 
           sta[5].n > 0 ? sta[5].n_frag*100.0 / sta[5].n : 0,
           sta[5].n > 0 ? sta[5].frag*1.0 / sta[5].n : 0);
      printf("\n");

      printf("---------------------------------------------------\n");
      sum.n = 0;
      sum.n_frag = 0;
      sum.frag = 0;
      for (i=0; i<6; i++)
      {
        sum.n += sta[i].n;
        sum.n_frag += sta[i].n_frag;
        sum.frag += sta[i].frag;
      }
      printf("    all %8lu    %8lu %8.2f  %11.2f\n",
             sum.n, sum.n_frag, 
             sum.n > 0 ? sum.n_frag*100.0 / sum.n : 0,
             sum.n > 0 ? sum.frag*1.0 / sum.n : 0);

      sum.n = 0;
      sum.n_frag = 0;
      sum.frag = 0;
      sum.n_blocks = 0;
      for (i=1; i<6; i++)
      {
        sum.n += sta[i].n;
        sum.n_frag += sta[i].n_frag;
        sum.frag += sta[i].frag;
	sum.n_blocks += sta[i].n_blocks;
      }
      printf("    > 1 %8lu    %8lu %8.2f  %11.2f\n",
             sum.n, sum.n_frag, 
             sum.n > 0 ? sum.n_frag*100.0 / sum.n : 0,
             sum.n > 0 ? sum.frag*1.0 / sum.n : 0);
      printf("Fragmentation index: %6.3f percent.\n", 
             (sum.frag - sum.n) * 100.0 / (sum.n_blocks - sum.n));

printf("frag: %lu  n: %lu  n_blocks: %lu  (n_frag: %lu)\n",
       sum.frag, sum.n, sum.n_blocks, sum.n_frag);

      printf("\n");
    }
  }
  return n_indir;
}   /* scan_directory() */



int main(int argc, char** argv)
{
  int i;
  char *c;
  char directory[PATH_MAX];
  unsigned int n_indir;
  struct statistics sum;      /* summary statistics */
  

  if (argc < 2) usage(argv[0]);
  for (i=1; i<argc; i++)
  {
    c = argv[i];
    if ('-' == *c)
    {
      c++;
      switch (*c)
      {
        case 'r' : recursive = 1; break;
        case 'd' : print_stats = 1; recursive = 1; break;
        case 'f' : print_files = 1; break;
        case 'b' : print_blocks = 1; print_files = 1; break;
        case 'x' : xdev = 1; break;
        default :  usage(argv[0]);
      }
    }
    else sprintf(directory, argv[i]);
  }

  for (i=0; i<6; i++)
  {
    global[i].n = 0;
    global[i].n_frag = 0;
    global[i].frag = 0;
    global[i].n_blocks = 0;
  }
  n_indir = scan_directory(directory);
  if (0 == n_indir)
  {
    printf("No files found in %s\n", directory);
    usage(argv[0]);
  }
  printf("\nSummary statistics for %s including subdirectories:\n", directory);
  printf(" blocks    files  fragmented  percent  fragm./file\n");
  printf("---------------------------------------------------\n");
  printf("      1 %8lu           0     0.00         %4.2f\n",
         global[0].n, global[0].n ? 1.0 : 0.0);
  printf("  <= 12 %8lu    %8lu %8.2f  %11.2f\n",
         global[1].n, global[1].n_frag, 
         global[1].n > 0 ? global[1].n_frag*100.0 / global[1].n : 0, 
         global[1].n > 0 ? global[1].frag*1.0 / global[1].n : 0);
  printf("<= %4u %8lu    %8lu %8.2f  %11.2f\n", 
         n_indir/2+12, global[2].n, global[2].n_frag, 
         global[2].n > 0 ? global[2].n_frag*100.0 / global[2].n : 0,
         global[2].n > 0 ? global[2].frag*1.0 / global[2].n : 0);
  printf("<= %4u %8lu    %8lu %8.2f  %11.2f\n", 
         n_indir+12, global[3].n, global[3].n_frag,
         global[3].n > 0 ? global[3].n_frag*100.0 / global[3].n : 0,
         global[3].n > 0 ? global[3].frag*1.0 / global[3].n : 0);
  printf("<= %4u %8lu    %8lu %8.2f  %11.2f\n",
         n_indir*4+12, global[4].n, global[4].n_frag,
         global[4].n > 0 ? global[4].n_frag*100.0 / global[4].n : 0,
         global[4].n > 0 ? global[4].frag*1.0 / global[4].n : 0);
  printf(" > %4u %8lu    %8lu %8.2f  %11.2f\n",
         n_indir*4+12, global[5].n, global[5].n_frag, 
         global[5].n > 0 ? global[5].n_frag*100.0 / global[5].n : 0,
         global[5].n > 0 ? global[5].frag*1.0 / global[5].n : 0);

  printf("---------------------------------------------------\n");
  sum.n = 0;
  sum.n_frag = 0;
  sum.frag = 0;
  for (i=0; i<6; i++)
  {
    sum.n += global[i].n;
    sum.n_frag += global[i].n_frag;
    sum.frag += global[i].frag;
  }
  printf("    all %8lu    %8lu %8.2f  %11.2f\n",
         sum.n, sum.n_frag, 
         sum.n > 0 ? sum.n_frag*100.0 / sum.n : 0,
         sum.n > 0 ? sum.frag*1.0 / sum.n : 0);

  sum.n = 0;
  sum.n_frag = 0;
  sum.frag = 0;
  sum.n_blocks = 0;
  for (i=1; i<6; i++)
  {
    sum.n += global[i].n;
    sum.n_frag += global[i].n_frag;
    sum.frag += global[i].frag;
    sum.n_blocks += global[i].n_blocks;
  }
  printf("    > 1 %8lu    %8lu %8.2f  %11.2f\n",
         sum.n, sum.n_frag, 
         sum.n > 0 ? sum.n_frag*100.0 / sum.n : 0,
         sum.n > 0 ? sum.frag*1.0 / sum.n : 0);

  printf("Fragmentation index: %6.3f percent.\n", 
         (sum.frag - sum.n) * 100.0 / (sum.n_blocks - sum.n));

printf("frag: %lu  n: %lu  n_blocks: %lu  (n_frag: %lu)\n",
       sum.frag, sum.n, sum.n_blocks, sum.n_frag);

  return 0;
}
